##################################################
# 動画から画像を切り出す
#  動画制御ウィンドウ
#
# Shin.Onda 2023/6//3 all rights reserved.
##################################################
import cv2
import numpy as np
import os
import datetime as dt
import tkinter as tk
import tkinter.filedialog as filedialog
import tkinter.messagebox as messagebox
import tkinter.simpledialog as simpledialog
import mousectrl
import adjustimage

# ウインドウ名
WINDOW_NAME = 'Video to pictures'

#動画ファイルタイプ
#(環境に依存します。必要に応じて変更してください)
MOVIE_FILETYPE = [("", "*.mp4"), ("", "*.avi"), ("", "*.wmv")]

#書き込み画像ファイル拡張子
#(環境に依存します。必要に応じて変更してください)
WRITE_EXT = 'jpg'

# 動画サイズ制限
# 1000000 frames = 30fps でおよそ９時間
MAX_MOVIE_FRAMES = 1000000
MIN_MOVIE_SIZE   = 16
# 画像調整クラスのインスタンス
adjustImage = adjustimage.AdjustImage()

class VideoToPicCtrl :
    #######################
    # メンバ変数
    #######################
    # 画像のデフォルト
    defaultDir = ""
    # 読み飛ばしフレーム数
    skipCount = 0
    # 正方形切り出しオプション
    cutSquareOption = False
    # 正方形に配置オプション
    squareOption = False
    # サイズ調整オプション
    adjustOption = False
    # サイズ調整ピクセル値
    adjustSize = 512
    # 拡大縮小オプション
    zoomOption = False
    # 拡大縮小サイズ（％最小）
    zoomMin = 50
    # 拡大縮小サイズ（％最大）
    zoomMax = 100

    # アクセサ
    def getDefaultDir(self) :
        return self.defaultDir
    def setDefaultDir(self, defaultDir) :
        self.defaultDir = defaultDir

    def getSkipCount(self) :
        return self.skipCount
    def setSkipCount(self, skipCount) :
        self.skipCount = skipCount

    def getCutSquareOption(self) :
        return self.cutSquareOption
    def setCutSquareOptin(self, cutSquareOption) :
        self.cutSquareOption = cutSquareOption

    def getSquareOption(self) :
        return self.squareOption
    def setSquareOptin(self, squareOption) :
        self.squareOption = squareOption

    def getAdjustOption(self) :
        return self.adjustOption
    def setAdjustOption(self, adjustOption) :
        self.adjustOption = adjustOption

    def getAdjustSize(self) :
        return self.adjustSize
    def setAdjustSize(self, adjustSize) :
        self.adjustSize = adjustSize

    def getZoomOption(self) :
        return self.zoomOption
    def setZoomOption(self, zoomOption) :
        self.zoomOption = zoomOption

    def getZoomMin(self) :
        return self.zoomMin
    def setZoomMin(self, zoomMin) :
        self.zoomMin = zoomMin
        
    def getZoomMax(self) :
        return self.zoomMax
    def setZoomMax(self, zoomMax) :
        self.zoomMax = zoomMax

    #######################
    # 動画ウインドウ制御
    #   fps         = fps
    #   totalFrames = 全フレーム数
    #   orgW       = 画像幅
    #   orgH       = 画像高さ
    # 戻り値 マウス制御オブジェクト
    #######################
    # 読込動画用のイメージウィンドウを開く
    def openMainWindow(self, fps, totalFrames, orgW, orgH, mouseCtrl) :
        # ウインドウを開く
        cv2.namedWindow(WINDOW_NAME, cv2.WINDOW_NORMAL)
        cv2.resizeWindow(WINDOW_NAME, int(orgW), int(orgH))
        cv2.setMouseCallback(WINDOW_NAME, mouseCtrl.mouseClick, param=(fps, totalFrames, orgW, orgH))

    #######################
    # フレームを表示
    #   frame     = 表示対象のフレーム
    #   orgW     = 画像幅
    #   orgH     = 画像高さ
    #   progress  = 動画のプログレス
    #   mouseCtrl = マウス制御オブジェクト
    #######################
    def showFrame(self, frame, orgW, orgH, progress, mouseCtrl) :
        frameWk = frame.copy()
        mouseCtrl.drawStartStop(frameWk, orgW, orgH)
        mouseCtrl.drawWriteButton(frameWk, orgW, orgH)
        mouseCtrl.drawFastButton(frameWk, orgW, orgH)
        mouseCtrl.drawBackButton(frameWk, orgW, orgH)
        mouseCtrl.drawSlide(frameWk, progress, orgW, orgH)
        cv2.imshow(WINDOW_NAME, frameWk)

    #######################
    # ファイル操作
    #######################
    #######################
    # 画像書き込み
    # cv2.imwrite で日本語パスが使用できない問題の対策
    #   filepath    = 書き込み先ファイル名パス名
    #   img         = 書き込むイメージ（フレーム）
    # 戻り値 True 正常、False 異常
    #######################
    def imwrite(self, filepath, img) :
        try:
            # 拡張子
            ext = os.path.splitext(filepath)[1]
            # エンコード
            ret, n = cv2.imencode(ext, img)
            if ret:
                # 画像をファイルに書き込む
                with open(filepath, mode='w+b') as f:
                    n.tofile(f)
            else:
                tk.Tk().withdraw()
                ret = messagebox.showinfo('異常', 'ファイル書き込みに失敗しました')
        except :
            tk.Tk().withdraw()
            ret = messagebox.showinfo('異常', 'ファイル書き込みに失敗しました')
            return False
        return True

    #######################
    # 画像のサイズ調整
    #   frame  = イメージ（フレーム）
    # 戻り値 frame 調整後のフレーム
    #######################
    def resizeFrame(self, frame) :
        # サイズ調整
        if self.getAdjustOption() :
            if self.getAdjustSize() is not None :
                resize = self.getAdjustSize()
                if resize > 1 :
                     h, w = frame.shape[:2]
                     h = int(h / w * resize)
                     frame = cv2.resize(frame, dsize=(resize, h))
        return frame

    #######################
    # フレームの書き込み
    #   orgFrame  = 元のイメージ（フレーム）
    #   filepath  = 書き込み先ファイル名パス名
    #   frameNo  = 現在のフレーム番号
    #   outputDir = 書き込み先フォルダ
    #   mouseCtrl = マウス制御オブジェクト
    #######################
    def writeFrame(self, orgFrame, filepath, frameNo, outputDir, mouseCtrl) :
        imgBack = 'R'

        # 中断チェック
        key = cv2.waitKey(1)
        if key == 27 : # ESCキー
            tk.Tk().withdraw()
            ret = messagebox.askyesno('確認', '書き込み処理を中止しますか？')
            if ret :
                # 書き込みモードを解除
                mouseCtrl.setWriteFlag(False)
                return

        # 拡大縮小
        print('Please wait...')
        frameList = []
        if self.getZoomOption() :
            min = self.getZoomMin()
            max = self.getZoomMax()
            if min < max :
                frameList = adjustImage.zoomSquare(orgFrame, imgBack, min, max, 10) 
            else :
                frameList.append(orgFrame)
        else :
            frameList.append(orgFrame)

        for i, frame in enumerate(frameList) :
            # 中断チェック
            key = cv2.waitKey(1)
            if key == 27 : # ESCキー
                tk.Tk().withdraw()
                ret = messagebox.askyesno('確認', '書き込み処理を中止しますか？')
                if ret :
                    # 書き込みモードを解除
                    mouseCtrl.setWriteFlag(False)
                    break

            # 元のファイル名に日時とフレームNoを付加する
            orgname = os.path.splitext(os.path.basename(filepath))[0]
            pathformat = '{0}_{1:06d}_{2}_{3}.{4}'
            writefilename = pathformat.format(orgname, (int(frameNo) % 1000000), 0, i, WRITE_EXT)
            outputPath = os.path.join(outputDir, writefilename)
            self.imwrite(outputPath, self.resizeFrame(frame))
            print(outputPath)

            # 正方形切り出し
            if self.getCutSquareOption() :
              frameList = adjustImage.cutSquare(frame)
              for j, f in enumerate(frameList) :
                  writefilename = pathformat.format(orgname, (int(frameNo) % 1000000), j+1, i, WRITE_EXT)
                  outputPath = os.path.join(outputDir, writefilename)
                  self.imwrite(outputPath, self.resizeFrame(frameList[j]))
                  print(outputPath)

            # 正方形に配置
            if self.getSquareOption() :
              frame2 = adjustImage.adjustSquare(frame, imgBack)
              writefilename = pathformat.format(orgname, (int(frameNo) % 1000000), 4, i, WRITE_EXT)
              outputPath = os.path.join(outputDir, writefilename)
              self.imwrite(outputPath, self.resizeFrame(frame2))
              print(outputPath)

    ###########################
    # 終了処理
    #   video  = オープンした動画
    ###########################
    def videoToPicEnd(self, video) :
        if video is not None :
            video.release()
        cv2.destroyAllWindows()

    ###########################
    # 動画読み込み・表示・ファイル書き込み
    ###########################
    def videoCtrl(self) :
        # 動画ファイル読み込みループ
        while True :
            # ファイル選択
            filepath = filedialog.askopenfilename(title="動画ファイル選択", filetypes=MOVIE_FILETYPE, initialdir=self.getDefaultDir())
            print(filepath)
            if len(filepath) == 0 :
                self.videoToPicEnd(None) # キャンセル
                print('Canceled.')
                return
            # デフォルトのフォルダ
            self.setDefaultDir(os.path.dirname(filepath))

            # ビデオファイルをオープン
            try :
                video = cv2.VideoCapture(filepath)
                if video.isOpened() == False :
                    print('Open error !')
                    print(filepath)
                    tk.Tk().withdraw()
                    messagebox.showinfo('異常', 'ファイルが読み込めません')
                    video.release()
                    continue  # ファイル再選択
            except :
                print('Open error !')
                print(filepath)
                tk.Tk().withdraw()
                messagebox.showinfo('異常', 'ファイル読み込みに失敗しました')
                video.release()
                continue  # ファイル再選択

            # ビデオのパラメータを獲得
            fps   = video.get(cv2.CAP_PROP_FPS)
            orgW = video.get(cv2.CAP_PROP_FRAME_WIDTH)
            orgH = video.get(cv2.CAP_PROP_FRAME_HEIGHT)
            totalFrames = video.get(cv2.CAP_PROP_FRAME_COUNT)
            watitime = int(1000.0/fps)
            print('Video properties fps:{0} w:{1} h:{2} '.format(fps, orgW, orgH))
            if orgW < MIN_MOVIE_SIZE or orgH < MIN_MOVIE_SIZE :
                tk.Tk().withdraw()
                messagebox.showinfo('確認', '申し訳ございません。動画が小さすぎるので処理できません。')
                video.release()
                continue  # ファイル再選択
            if totalFrames > MAX_MOVIE_FRAMES :
                tk.Tk().withdraw()
                messagebox.showinfo('確認', '申し訳ございません。動画が長すぎるので処理できません。')
                video.release()
                continue  # ファイル再選択

            # フレーム読み込み
            retval, frame = video.read()
            if frame is None :
                print('Readerror!!')
                tk.Tk().withdraw()
                messagebox.showinfo('異常', 'ファイルが読み込めません')
                video.release()
                continue  # ファイル再選択
            # 読み込み正常ならループ終了
            break;

        # 書き込み先フォルダ選択ループ
        while True :
            # フォルダ選択
            outputDir = filedialog.askdirectory(title="書込先フォルダ選択", initialdir=self.getDefaultDir())
            print(outputDir)
            if len(outputDir) == 0 :
                print('Canceled.')
                self.videoToPicEnd(video) # キャンセル
                return
            # 出力フォルダ作成
            outputDir = os.path.join(outputDir, os.path.splitext(os.path.basename(filepath))[0])
            tk.Tk().withdraw()
            if os.path.isdir(outputDir) :
                msg = 'フォルダ[{0}]に書き込みます。既存ファイルがあれば上書きされます。よろしいですか？'.format(outputDir)
                ret = messagebox.askyesno('確認', msg)
                if not ret :
                    continue  # フォルダ再選択
            else :
                msg = 'フォルダ[{0}]を作成します。よろしいですか？'.format(outputDir)
                ret = messagebox.askyesno('確認', msg)
                if not ret :
                    continue  # フォルダ再選択
                os.makedirs(outputDir)

            # 読み込み正常ならループ終了
            break
        
        # マウス制御クラスのインスタンス
        mouseCtrl = mousectrl.MouseCtrl()
        # 読込動画用のイメージウィンドウを開く
        self.openMainWindow(fps, totalFrames, orgW, orgH, mouseCtrl)

        # 動画制御ループ
        while True :
            # ウインドウ状態チェック
            ret = cv2.getWindowProperty(WINDOW_NAME, cv2.WND_PROP_VISIBLE)
            if not ret :
                tk.Tk().withdraw()
                ret = messagebox.askyesno('確認', 'この動画の処理を終了しますか？')
                if ret :
                    break;  # 閉じられたら終了
                else :
                    # いいえで再表示
                    self.openMainWindow(fps, totalFrames, orgW, orgH, mouseCtrl)

            # フレームを表示
            if mouseCtrl.getFrameNo() >= (totalFrames - 1) :
                progress = 1
            else :
                progress = mouseCtrl.getFrameNo() / (totalFrames - 1)
            self.showFrame(frame, orgW, orgH, progress, mouseCtrl)

            # FPSに基づき一時停止、停止(ESC)キー入力なら終了
            key = cv2.waitKey(watitime)
            if key == 27 : # ESCキー
                if mouseCtrl.getWriteFlag() :
                    tk.Tk().withdraw()
                    ret = messagebox.askyesno('確認', '書き込みモードを終了しますか？')
                    if ret :
                        mouseCtrl.setWriteFlag(False)
                else :
                    tk.Tk().withdraw()
                    ret = messagebox.askyesno('確認', 'この動画の処理を終了しますか？')
                    if ret :
                        break
        
            # 書き込みモードなら１フレーム書き込み
            if mouseCtrl.getWriteFlag() :
                if mouseCtrl.getMoveFlag() :
                    # 書き込み中フレームを表示
                    mouseCtrl.setProcessFlag(True)
                    self.showFrame(frame, orgW, orgH, (mouseCtrl.getFrameNo()/totalFrames), mouseCtrl)
                    # フレームが移動している場合、読み飛ばしフレーム数分読み飛ばして書き込む
                    if (mouseCtrl.getWriteCounter() % (self.getSkipCount() + 1)) == 0 :
                        self.writeFrame(frame, filepath, mouseCtrl.getFrameNo(), outputDir, mouseCtrl)
                    mouseCtrl.setWriteCounter(mouseCtrl.getWriteCounter() + 1)
                    # 書き込み終了フレームを表示
                    mouseCtrl.setProcessFlag(False)
                    self.showFrame(frame, orgW, orgH, (mouseCtrl.getFrameNo()/totalFrames), mouseCtrl)

            # 移動中なら次のフレームへ
            if mouseCtrl.getMoveFlag() :
                if (totalFrames - 1) > mouseCtrl.getFrameNo() :
                    # 次のフレームへ
                    mouseCtrl.setFrameNo(mouseCtrl.getFrameNo() + 1)
                else :
                    # 読み込み終了、停止
                    mouseCtrl.setMoveFlag(False)
        
            # フレーム読み込み
            video.set(cv2.CAP_PROP_POS_FRAMES, mouseCtrl.getFrameNo())
            retval, frameWk = video.read()
            if frameWk is None :
                # 次がなければフレームNoを戻す
                mouseCtrl.setFrameNo(mouseCtrl.getFrameNo() - 1)
                # 読み込み終了、停止
                mouseCtrl.setMoveFlag(False)
            else :
                frame = frameWk
        
        ## -- キーループ 終了 --
        #終了処理
        self.videoToPicEnd(video)


